//============================================================================
//
// NTCrash
//
// Copyright (C) 1996 Mark Russinovich
//
// Discovers holes in win32k parameter validation.
//
//============================================================================

#include <windows.h>
#include <stdio.h>
#include <process.h>
#include <stdio.h>
#include <time.h>


#define TESTING 0

// threads only spawned when logging is off
#define NUMADDTHREADS 10

// log file handle
HANDLE	logfile;

// buffer that will go to the log file
char	logrec[4096];

//
// bad call array - pre-initialized with calls that we can't get past
//
int totbad = 0;
int badcalls[512];

// log the calls to a file?
BOOL	logcalls = FALSE;

// do ntoskrnl?
BOOL	ntoskrnl = FALSE;


//============================================================================
//
// scanlog
//
// Reads the bad.log file to obtain system calls that are known to fail,
// and then appends the last call from crash.log, since that was the last
// executed before the last failure.
//
//============================================================================
void scanlog()
{
	FILE	*fp;
	int		badnumber, bw, len;
	char	badline[1024], nextline[1024];
	HANDLE	badfile;

	// read bad call file 
	fp = fopen( "bad.log", "r" );
	if( fp ) {
		printf("Scanning bad.log\n");
		while( fgets( badline, 1024, fp ) ) {
			if( strlen( badline ) > 2 ) {
				sscanf( badline, "0x%x", &badnumber );
				badcalls[ totbad++] = badnumber;
				printf("  Excluding 0x%x\n", badnumber );
			}
		}
		fclose( fp );
	}

	// now get the latest from the crash log
	fp = fopen( "crash.log" , "r" );
	if( fp ) {

		// scan the file to get the last line recorded
		badline[0] = 0;
		printf("Scanning crash.log\n");
		while( fgets( nextline, 1024, fp )) {
			if( strlen( nextline ) > 2 )
				strcpy(badline, nextline);
		}
		fclose( fp );

		// were there any records in the crash log?
		if( badline[0] ) {

			// extract the failed call number
			sscanf( badline, "0x%x", &badnumber );
			badcalls[ totbad++] = badnumber;
			printf("  Excluding 0x%x\n", badnumber );
			
			// add it to the bad call file
			badfile = CreateFile("bad.log", GENERIC_WRITE, 0, NULL, 
					OPEN_ALWAYS, FILE_FLAG_WRITE_THROUGH, 0 );

			// fix the carriage return
			len = strlen( badline );
			badline[ len-1 ] = '\r';
			badline[ len   ] = '\n';
			badline[ len+1 ] = 0;
			SetFilePointer( badfile, 0, 0, FILE_END );
			WriteFile( badfile, badline, strlen(badline), &bw, 0);
			FlushFileBuffers(badfile); // just to be safe
			CloseHandle( badfile );
		}
	}	


	// create a new log
	DeleteFile("crash.log");
	logfile = CreateFile("crash.log", GENERIC_WRITE, 0, NULL, 
			CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH, 0 );
	CloseHandle(logfile);
}


//============================================================================
//
// logrecord
//
// Synchronously writes an entry into the log file.
//
//============================================================================
void logrecord( int callnumber, long *stackptr )
{
	char	param[32];
	int		br, i;

	// just to be safe
	memset( logrec, 0, 4096);
	sprintf( logrec, "0x%x => ( ", callnumber );

	for( i = 0; i < 15; i++ ) {
		sprintf( param, "0x%x, ", stackptr[i] );
		strcat( logrec, param );
	}
	strcat( logrec, " )\r\n" );
#if TRUNCATE
	logfile = CreateFile("crash.log", GENERIC_WRITE, 0, NULL, 
		CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH, 0 );
#else
	logfile = CreateFile("crash.log", GENERIC_WRITE, 0, NULL, 
		OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH, 0 );
	SetFilePointer( logfile, 0, 0, FILE_END );
#endif
	WriteFile( logfile, logrec, strlen(logrec), &br, NULL);
	FlushFileBuffers(logfile); // just to be safe
	CloseHandle(logfile);
}


//============================================================================
//
// threadloop
//
// Just spin while firing random calls pointing at garbage into Win32k.sys
//
//============================================================================
void threadloop( void *param )
{
	int		loop = 0;
	int		i;
	int		callnumber;
	int		thrdnum = (int) param;
	long	stack[100];
	long	*stackptr;

	while(1) {

		// create garbage parameters once every 1000 trys
		if( !(++loop % 1000)) {
			printf(".");

			// play with the seed
			srand((unsigned)time( NULL ));
			fflush(stdout);
		}


		// go for it
		__try {

			// generate a call number
makecall:
			if( ntoskrnl )
				callnumber = (rand() % 0xD2);
			else
				callnumber = (rand() % 0x1f1) + 0x1000;

			// Win32k 0x11eb causes just this process to hang
			if (!ntoskrnl && callnumber == 0x11eb ) callnumber++;

			if( logcalls ) {

				// make sure its not a call we know fails
				for( i = 0; i < totbad; i++ ) {
					if( badcalls[i] == callnumber ) goto makecall;
				}
			}

			// push params
			for( i = 0 ; i < 40; i++) 
				stack[i] = rand() << 16 | rand();
			
			// save record to log file
			if( logcalls ) logrecord( callnumber, stack );
#if !TESTING
			// do the systrap
			stackptr = stack;
			_asm mov eax, callnumber;
			_asm mov edx, stackptr;
			_asm int 2eh;
#endif

		} __except( EXCEPTION_EXECUTE_HANDLER ) {

		} 
	}
}

//============================================================================
//
// usage
//
//============================================================================
void usage( char *name )
{
	printf("usage: %s [-l] [-n]\n"
			"-l		Log to stable file\n"
			"-n		Exercise NTOSKRNL\n", name );
	exit(1);
}


//============================================================================
//
// main
//
// If logging, saves info to a log file. If not, fire off a bunch of threads
// to crash the system real quick-like.
//
//============================================================================
void main( int argc, char *argv[])
{
	int i;
	printf("Crashme V1.0\n");
	printf("Copyright (C) 1996 Mark Russinovich\n");

	for( i = 1; i < argc; i ++ ) {
		if( argv[i][0] != '-' )
			usage( argv[0] );
		switch( argv[i][1] ) {
		case 'l':
		case 'L':
			logcalls = TRUE;
			printf("Logging calls to crash.log\n" );
			break;
		case 'n':
		case 'N':
			printf("Exercising NTOSKRNL\n" );
			ntoskrnl = TRUE;
			break;
		default:
			usage(argv[0]);
		}
	}

	if( !ntoskrnl )
		printf("Exercising WIN32K\n");
		
	
	// Do logging setup
	if( logcalls ) {
		scanlog();
	}

	printf("Running stress test. Press Ctrl-C to quit.\n");

	// spawn multiple thread if not logging
	if( !logcalls ) {
		for( i = 0; i < NUMADDTHREADS;i ++ ) 
			_beginthread( threadloop, 0x4000, (void *) (i+1) );
	}
	threadloop( (void *) 0 );
}
